"""Tools for autotuning algorithm parameters and caching the results.
Note that this is intended to apply only to parameters that affect
performance, such as work-group sizes, and not the outputs.

The design is that each computation class that implements autotuning will
provide a class method (typically called `autotune`) that takes a context and
user-supplied parameters (related to the problem rather than the
implementation), and returns a dictionary of values. A decorator is applied to
the autotuning method to cause the results to be cached.

The class may define an autotune_version class attribute to version the table.

At present, caching is implemented in an sqlite3 database. There is a table
corresponding to each autotuning method. The table has the following columns:

- `device_name`: string, name for the compute device
- `device_platform`: string, name for the compute device's platform
- `device_version`: version string for the driver
- `arg_*`: for each parameter passed to the autotuner
- `value_*`: for each value returned from the autotuner

The database is stored in the user cache directory.
"""

from __future__ import division, print_function, absolute_import

import itertools
import inspect
import os
import os.path
import enum
import time
import logging
import multiprocessing
import concurrent.futures

import appdirs
import sqlite3
import numpy as np
from decorator import decorator


_logger = logging.getLogger(__name__)


def adapt_value(value):
    """Converts `value` to a type that can be used in sqlite3. This is
    not done through the sqlite3 adapter interface, because that is global
    rather than per-connection. This also only applies to lookup keys,
    not results, because it is not a symmetric relationship.
    """
    if isinstance(value, type) or isinstance(value, np.dtype):
        return repr(value)
    elif isinstance(value, enum.Enum):
        return value.name
    return value


def _db_keys(fn, args, kwargs):
    """Uses the arguments passed to an autotuning function (see
    :function:`autotuner`) to generate a database key.
    """
    argspec = inspect.getargspec(fn)
    # Extract the arguments passed to the wrapped function, by name
    named_args = dict(kwargs)
    for i in range(2, len(args)):
        named_args[argspec.args[i]] = args[i]
    keys = dict([('arg_' + key, adapt_value(value)) for (key, value) in named_args.items()])

    # Add information about the device
    device = args[1].device
    keys['device_name'] = device.name
    keys['device_platform'] = device.platform_name
    keys['device_version'] = device.driver_version
    return keys


def _query(conn, tablename, keys):
    """Fetch a cached record from the database. If the record is not found,
    it will return None.

    Parameters
    ----------
    conn : sqlite3.Connection
        Database connection
    tablename : str
        Name of the table to query
    keys : mapping
        Keys and values for the query
    """
    try:
        query = 'SELECT * FROM {0} WHERE'.format(tablename)
        query_args = []
        first = True
        for key, value in keys.items():
            if not first:
                query += ' AND'
            first = False
            query += ' {0}=?'.format(key)
            query_args.append(value)
        cursor = conn.cursor()
        cursor.execute(query, query_args)
        row = cursor.fetchone()
        if row is not None:
            ans = {}
            for colname in row.keys():
                if colname.startswith('value_'):
                    # Truncate the 'value_' prefix
                    ans[colname[6:]] = row[colname]
            return ans
    except sqlite3.Error:
        # This could happen if the table does not exist yet
        _logger.debug("Query '%s' failed", query, exc_info=True)
        pass
    return None


def _create_table(conn, tablename, keys, values):
    command = 'CREATE TABLE IF NOT EXISTS {0} ('.format(tablename)
    for name in itertools.chain(keys.keys(), values.keys()):
        command += name + ' NOT NULL, '
    command += 'PRIMARY KEY ({0}) ON CONFLICT REPLACE)'.format(', '.join(keys.keys()))
    conn.execute(command)


def _save(conn, tablename, keys, values):
    """Write cached result into the table, creating it if it does not already
    exist.
    """
    _create_table(conn, tablename, keys, values)
    # Combine all fields
    entries = dict(keys)
    entries.update(values)
    command = 'INSERT OR REPLACE INTO {0}({1}) VALUES ({2})'.format(
        tablename,
        ', '.join(entries.keys()),
        ', '.join(['?' for x in entries.keys()]))
    # Start transaction
    with conn:
        cursor = conn.cursor()
        cursor.execute(command, list(entries.values()))


def _open_db():
    cache_dir = appdirs.user_cache_dir('katsdpsigproc', 'ska-sa')
    try:
        os.makedirs(cache_dir)
    except OSError:
        # This happens if the directory already exists. If we failed
        # to create it, the database open will fail.
        pass

    cache_file = os.path.join(cache_dir, 'tuning.db')
    conn = sqlite3.connect(cache_file)
    return conn


def _close_db(conn):
    """Close a database. This is split into a separate function for the benefit
    of testing, because the close member of the object itself is read-only and
    hence cannot be patched."""
    conn.close()


def autotuner_impl(test, fn, *args, **kwargs):
    """Implementation of :func:`autotuner`. It is split into a separate
    function so that mocks can patch it."""
    cls = args[0]
    classname = '{0}.{1}.{2}'.format(cls.__module__, cls.__name__, fn.__name__)
    tablename = classname.replace('.', '_') + \
        '__' + str(getattr(cls, 'autotune_version', 0))
    keys = _db_keys(fn, args, kwargs)

    conn = _open_db()
    conn.row_factory = sqlite3.Row
    try:
        ans = _query(conn, tablename, keys)
        if ans is None:
            # Nothing found in the database, so we need to tune now
            _logger.info('Performing autotuning for %s with key %s', classname, keys)
            ans = fn(*args, **kwargs)
            values = dict([('value_' + key, value) for (key, value) in ans.items()])
            _save(conn, tablename, keys, values)
        else:
            _logger.debug('Autotuning cache hit for %s with key %s', classname, keys)
    finally:
        _close_db(conn)
    return ans


def autotuner(test):
    r"""Decorator that marks a function as an autotuning function and caches
    the result. The function must take a class and a context as the first
    two arguments. The remaining arguments form a cache key, along with
    properties of the device and the name of the function.

    Every argument to the function must have a name, which implies that the
    \*args construct may not be used.

    Parameters
    ----------
    test : dictionary
        A value that will be returned by :func:`stub_autotuner`.
    """
    @decorator
    def autotuner(fn, *args, **kwargs):
        r"""Decorator that marks a function as an autotuning function and caches
        the result. The function must take a class and a context as the first
        two arguments. The remaining arguments form a cache key, along with
        properties of the device and the name of the function.

        Every argument to the function must have a name, which implies that the
        \*args construct may not be used.
        """
        return autotuner_impl(test, fn, *args, **kwargs)
    return autotuner


def force_autotuner(test, fn, *args, **kwargs):
    """Drop-in replacement for :func:`autotuner_impl` that does not do any
    caching. It is intended to be used with a mocking framework.
    """
    return fn(*args, **kwargs)


def stub_autotuner(test, fn, *args, **kwargs):
    """Drop-in replacement for :func:`autotuner_impl` that does not do
    any tuning, but instead returns the provided value. It is intended to be
    used with a mocking framework."""
    return test


def make_measure(queue, function):
    """Generates a measurement function that can be returned by the
    function passed to :func:`autotune`. It calls `function`
    (with no arguments) the appropriate number of times and returns
    the averaged elapsed time as measured by `queue`."""
    def measure(iters):
        queue.start_tuning()
        for i in range(iters):
            function()
        return queue.stop_tuning() / iters
    return measure


def autotune(generate, time_limit=0.1, threads=None, **kwargs):
    """Run a number of tuning experiments and find the optimal combination
    of parameters.

    Each argument is a iterable. The `generate` function is passed each
    element of the Cartesian product (by keyword), and returns a callable.
    This callable is passed an iteration count, and returns a score: lower is
    better. If either `generate` or the function it returns raises an
    exception, it is suppressed. Returns a dictionary with the best combination
    of values.

    Instead of a callable, the generate function may return `None` to skip a
    configuration that it knows will be unsuitable. Throwing an exception has
    essentially the same effect (and is used in code written before returning
    `None` was allowed), but returning `None` is preferred since an exception
    just clutters the log.

    The scoring function should not do a warmup pass: that is handled by this
    function.

    Parameters
    ----------
    generate : callable
        function that creates a scoring function
    time_limit : float
        amount of time to spend testing each configuration (excluding setup time)
    threads : int, optional
        number of parallel calls to `generate` to make at a time (defaults to
        CPU count)

    Raises
    ------
    Exception : if every combination throws an exception, the last exception
        is re-raised.
    ValueError : if the Cartesian product is empty
    """
    opts = itertools.product(*kwargs.values())
    best = None
    best_score = None
    exc = None
    if threads is None:
        try:
            threads = multiprocessing.cpu_count()
        except NotImplementedError:
            threads = 1
    with concurrent.futures.ThreadPoolExecutor(threads) as thread_pool:
        while True:
            batch = list(itertools.islice(opts, threads))
            if not batch:
                break
            batch_keywords = [dict(zip(kwargs.keys(), opt)) for opt in batch]
            futures = [thread_pool.submit(generate, **keywords) for keywords in batch_keywords]
            concurrent.futures.wait(futures)
            for keywords, future in zip(batch_keywords, futures):
                try:
                    measure = future.result()
                    if measure is None:
                        continue
                    # Do a warmup pass
                    measure(1)
                    # Do an initial timing pass
                    start = time.time()
                    measure(1)
                    # Take a max to prevent divide-by-zero in very fast cases
                    elapsed = max(time.time() - start, 1e-4)
                    # Guess how many iterations we can do in the allotted time
                    iters = max(3, int(time_limit / elapsed))
                    score = measure(iters)
                    _logger.debug("Configuration %s scored %f in %d iterations",
                                  keywords, score, iters)
                    if best_score is None or score < best_score:
                        best = keywords
                        best_score = score
                except Exception as e:
                    exc = e
                    _logger.debug("Caught exception while testing configuration %s",
                                  keywords, exc_info=True)
    if best is None:
        if exc is not None:
            raise exc
        else:
            raise ValueError('No options to test')
    return best



"""Abstraction layer over PyOpenCL to present an interface that is common
between CUDA and OpenCL. This is only the subset that is needed so far. New
functionality can be added, but should be kept in sync with
:mod:`katsdpsigproc.cuda`.
"""

import pyopencl
import pyopencl.array
import numpy as np


class _DummyArray(object):
    """Trivial dummy for :class:`pyopencl.array.Array`, that just has
    a `data` attribute.
    """
    def __init__(self, data):
        self.data = data


class _PinnedAMD(np.ndarray):
    """Pinned memory abstraction for AMD GPUs. Copies to the device
    are done by unmapping the buffer, enqueuing the copy, and remapping
    the buffer. This is based on AMD's optimization guide.
    """
    def __new__(cls, context, queue, shape, dtype):
        dtype = np.dtype(dtype)
        n_bytes = int(np.product(shape)) * dtype.itemsize
        # Do not add READ or WRITE to the flags: doing so seems to cause AMD
        # drivers to allocate GPU memory.
        buffer = pyopencl.Buffer(
            context,
            pyopencl.mem_flags.ALLOC_HOST_PTR,
            n_bytes)
        array = pyopencl.enqueue_map_buffer(
            queue,
            buffer,
            pyopencl.map_flags.READ | pyopencl.map_flags.WRITE,
            0, shape, dtype)[0]
        mapping = array.base
        array = array.view(_PinnedAMD)
        array._buffer = buffer
        array._mapping = mapping
        array._array = _DummyArray(buffer)
        return array

    def _enqueue_transfer(self, queue, blocking, transfer):
        if not hasattr(self, '_buffer'):
            raise ValueError('cannot copy to/from a _PinnedAMD view')
        self._mapping.release(queue._pyopencl_command_queue)
        transfer()
        # The shape and dtype don't matter here as long as we map the entire buffer,
        # because we're going to throw away the array wrapper and just keep the
        # MemoryMap object.
        array = pyopencl.enqueue_map_buffer(
            queue._pyopencl_command_queue,
            self._buffer, pyopencl.map_flags.READ | pyopencl.map_flags.WRITE,
            0, (self._buffer.size,), np.uint8, is_blocking=blocking)[0]
        self._mapping = array.base

    def enqueue_write_buffer(self, queue, buffer, blocking):
        """Enqueue a copy from the host buffer to device memory. The host
        memory must not be touched while the copy is in progress.

        Parameters
        ----------
        queue : :class:`CommandQueue`
            Queue for the copy
        buffer : `pyopencl.array.Array`
            Target buffer
        blocking : bool
            If `True`, wait for the copy to complete before returning
        """
        def transfer():
            queue.enqueue_copy_buffer(self._array, buffer)
        self._enqueue_transfer(queue, blocking, transfer)

    def enqueue_read_buffer(self, queue, buffer, blocking):
        """Enqueue a copy to the host buffer from device memory. The host
        memory must not be touched while the copy is in progress.

        Parameters
        ----------
        queue : :class:`CommandQueue`
            Queue for the copy
        buffer : `pyopencl.array.Array`
            Source buffer
        blocking : bool
            If `True`, wait for the copy to complete before returning
        """
        def transfer():
            queue.enqueue_copy_buffer(buffer, self._array)
        self._enqueue_transfer(queue, blocking, transfer)

    def enqueue_write_buffer_rect(
            self, queue, buffer, buffer_origin, data_origin, shape,
            buffer_strides, data_strides, blocking=True):
        def transfer():
            queue.enqueue_copy_buffer_rect(
                self._array, buffer, data_origin, buffer_origin, shape,
                data_strides, buffer_strides)
        self._enqueue_transfer(queue, blocking, transfer)

    def enqueue_read_buffer_rect(
            self, queue, buffer, buffer_origin, data_origin, shape,
            buffer_strides, data_strides, blocking=True):
        def transfer():
            queue.enqueue_copy_buffer_rect(
                buffer, self._array, buffer_origin, data_origin, shape,
                buffer_strides, data_strides)
        self._enqueue_transfer(queue, blocking, transfer)


class Program(object):
    """Abstraction of a program object"""
    def __init__(self, pyopencl_program):
        self._pyopencl_program = pyopencl_program

    def get_kernel(self, name):
        """Create a new kernel

        Parameters
        ----------
        name : str
            Name of the kernel function
        """
        return Kernel(self, name)


class Kernel(object):
    """Abstraction of a kernel object. The object can be enqueued using
    :meth:`CommandQueue.enqueue_kernel`.

    The recommended way to create this object is via
    :meth:`Program.get_kernel`.
    """
    def __init__(self, program, name):
        self._pyopencl_kernel = pyopencl.Kernel(program._pyopencl_program, name)


class Event(object):
    """Abstraction of an event. This is more akin to a CUDA event than an
    OpenCL event, in that it is a marker in a command queue rather than
    associated with a specific command.
    """
    def __init__(self, pyopencl_event):
        self._pyopencl_event = pyopencl_event

    def wait(self):
        """Block until the event has completed"""
        self._pyopencl_event.wait()

    def time_since(self, prior_event):
        """Return the time in seconds from `prior_event` to self. Unlike the
        PyCUDA method of the same name, this will wait for the events to
        complete if they have not already.
        """
        prior_event.wait()
        self.wait()
        return 1e-9 * (self._pyopencl_event.profile.start
                       - prior_event._pyopencl_event.profile.end)

    def time_till(self, next_event):
        """Return the time in seconds from this event to `next_event`. See
        :meth:`time_since`.
        """
        return next_event.time_since(self)


class Device(object):
    """Abstraction of an OpenCL device"""
    def __init__(self, pyopencl_device):
        self._pyopencl_device = pyopencl_device

    def make_context(self):
        """Create a new context associated with this device"""
        return Context(pyopencl.Context([self._pyopencl_device]))

    @property
    def name(self):
        """Return human-readable name for the device"""
        return self._pyopencl_device.name

    @property
    def platform_name(self):
        """Return human-readable name for the platform owning the device"""
        return self._pyopencl_device.platform.name

    @property
    def driver_version(self):
        """Return human-readable name for the driver version"""
        return self._pyopencl_device.driver_version

    @property
    def is_cuda(self):
        """Whether the device is a CUDA device"""
        return False

    @property
    def is_gpu(self):
        """Whether device is a GPU"""
        return self._pyopencl_device.type & pyopencl.device_type.GPU

    @property
    def is_accelerator(self):
        """Whether device is an accelerator (as defined by OpenCL device
        types)"""
        return self._pyopencl_device.type & pyopencl.device_type.ACCELERATOR

    @property
    def is_cpu(self):
        """Whether the device is a CPU"""
        return self._pyopencl_device.type & pyopencl.device_type.CPU

    @property
    def simd_group_size(self):
        """The number of workitems that run in lock-step. This must only
        be used to tune performance parameters; there are no guarantees about
        memory coherency, forward progress etc.
        """
        return pyopencl.characterize.get_simd_group_size(self._pyopencl_device, 4)

    @classmethod
    def _get_platforms(cls):
        """Return all platforms"""
        try:
            return pyopencl.get_platforms()
        except pyopencl.LogicError:
            # OpenCL considers it an error if there are no platforms
            # available.
            return []

    @classmethod
    def get_devices(cls):
        """Return a list of all devices on all platforms"""
        ans = []
        for platform in cls._get_platforms():
            for device in platform.get_devices():
                ans.append(Device(device))
        return ans

    @classmethod
    def get_devices_by_platform(cls):
        """Return a list of all devices, with a sub-list per platform."""
        ans = []
        for platform in cls._get_platforms():
            ans.append([Device(device) for device in platform.get_devices()])
        return ans


class Context(object):
    """Abstraction of an OpenCL context"""
    def __init__(self, pyopencl_context):
        self._pyopencl_context = pyopencl_context
        device = pyopencl_context.devices[0]
        self._internal_queue = pyopencl.CommandQueue(pyopencl_context, device)
        self._force_pinned_amd = False   # Used for unit tests

    @property
    def device(self):
        """Return the device associated with the context (or the first device,
        if there are multiple)."""
        return Device(self._pyopencl_context.devices[0])

    def compile(self, source, extra_flags=None):
        """Build a program object from source

        Parameters
        ----------
        source : str
            OpenCL source code
        extra_flags : list
            OpenCL-specific parameters to pass to the compiler
        """
        # source is passed through str because it might arrive as Unicode,
        # triggering a warning
        program = pyopencl.Program(self._pyopencl_context, str(source))
        program.build(extra_flags)
        return Program(program)

    def allocate_raw(self, n_bytes):
        """Create an untyped buffer on the device."""
        return pyopencl.Buffer(self._pyopencl_context, pyopencl.mem_flags.READ_WRITE, n_bytes)

    def allocate(self, shape, dtype, raw=None):
        """Create a typed buffer on the device.

        Parameters
        ----------
        shape : tuple
            Shape for the array
        dtype : numpy dtype
            Type for the data
        raw : PyOpenCL buffer, optional
            Memory backing the array (automatically allocated if `None`
        """
        return pyopencl.array.Array(self._pyopencl_context, shape, dtype, data=raw)

    def allocate_pinned(self, shape, dtype):
        """Create a buffer in host memory that can be efficiently copied
        to and from the device.

        Parameters
        ----------
        shape : tuple
            Shape for the array
        dtype : numpy dtype
            Type for the data
        """
        device = self._pyopencl_context.devices[0]
        if (self._force_pinned_amd
            or ((device.type & pyopencl.device_type.GPU)
                and device.platform.name == 'AMD Accelerated Parallel Processing')):
            return _PinnedAMD(self._pyopencl_context, self._internal_queue, shape, dtype)
        elif device.platform.name == 'NVIDIA CUDA':
            # Based on NVIDIA's recommendation: create a device buffer and
            # leave it mapped permanently.
            n_bytes = int(np.product(shape)) * dtype.itemsize
            buf = pyopencl.Buffer(
                self._pyopencl_context,
                pyopencl.mem_flags.ALLOC_HOST_PTR | pyopencl.mem_flags.READ_ONLY,
                n_bytes)
            return pyopencl.enqueue_map_buffer(
                self._internal_queue,
                buf,
                pyopencl.map_flags.READ | pyopencl.map_flags.WRITE,
                0, shape, dtype)[0]
        else:
            return np.empty(shape, dtype)

    def allocate_svm_raw(self, n_bytes):
        raise NotImplementedError("PyOpenCL does not support OpenCL Shared Virtual Memory")

    def allocate_svm(self, shape, dtype, raw=None):
        raise NotImplementedError("PyOpenCL does not support OpenCL Shared Virtual Memory")

    def create_command_queue(self, profile=False):
        """Create a new command queue associated with this context

        Parameters
        ----------
        profile : boolean
            If true, the command queue will support timing kernels
        """
        return CommandQueue(self, profile=profile)

    def create_tuning_command_queue(self):
        """Create a new command queue for doing autotuning"""
        return TuningCommandQueue(self)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False


class CommandQueue(object):
    """Abstraction of a command queue. If no existing command queue is passed
    to the constructor, a new one is created.

    Parameters
    ----------
    context : :class:`Context`
        Context owning the queue
    pyopencl_command_queue : :class:`pyopencl.CommandQueue` (optional)
        Existing command queue to wrap
    profile : boolean
        If true and `pyopencl_command_queue` is omitted, enabling profiling
        (timing) on the queue.
    """
    def __init__(self, context, pyopencl_command_queue=None, profile=False):
        self.context = context
        if pyopencl_command_queue is None:
            pyopencl_device = context._pyopencl_context.devices[0]
            properties = 0
            if profile:
                properties |= pyopencl.command_queue_properties.PROFILING_ENABLE
            pyopencl_command_queue = pyopencl.CommandQueue(
                context._pyopencl_context, pyopencl_device,
                properties)
        self._pyopencl_command_queue = pyopencl_command_queue

    def enqueue_read_buffer(self, buffer, data, blocking=True):
        """Copy data from the device to the host. Only whole-buffer copies are
        supported, and the shape and type must match. In general, one should use
        the convenience functions in :class:`accel.DeviceArray`.

        Parameters
        ----------
        buffer : `pyopencl.array.Array`
            Source
        data : array-like
            Target
        blocking : boolean (optional)
            If true (default) the call blocks until the copy is complete
        """
        if hasattr(data, 'enqueue_read_buffer'):
            data.enqueue_read_buffer(self, buffer, blocking)
        else:
            buffer.get(ary=data, queue=self._pyopencl_command_queue, async_=not blocking)

    def enqueue_write_buffer(self, buffer, data, blocking=True):
        """Copy data from the host to the device. Only whole-buffer copies are
        supported, and the shape and type must match. In general, one should
        use the convenience functions in :class:`accel.DeviceArray`.

        Parameters
        ----------
        buffer : :class:`pyopencl.array.Array`
            Target
        data : array-like
            Source
        blocking : boolean (optional)
            If true (default), the call blocks until the source has been fully
            read (it has no necessarily reached the device).
        """
        if hasattr(data, 'enqueue_write_buffer'):
            data.enqueue_write_buffer(self, buffer, blocking)
        else:
            buffer.set(ary=data, queue=self._pyopencl_command_queue, async_=not blocking)

    def enqueue_copy_buffer(
            self, src_buffer, dest_buffer):
        """Copy one buffer to another.

        Parameters
        ----------
        src_buffer,dest_buffer : :class:`pyopencl.array.Array`
            Source and destination buffers
        """
        pyopencl.enqueue_copy(
            self._pyopencl_command_queue,
            src=src_buffer.data, dest=dest_buffer.data)

    def enqueue_copy_buffer_rect(
            self, src_buffer, dest_buffer, src_origin, dest_origin,
            shape, src_strides, dest_strides):
        """Copy a subregion of one buffer to another. This is a low-level
        interface that ignores the shape, strides etc of the buffers, and
        treats them as byte arrays. It also only supports 3 or fewer
        dimensions. Use
        :py:meth:`~katsdpsigproc.accel.DeviceArray.copy_region` for a
        high-level interface.

        Parameters
        ----------
        src_buffer,dest_buffer : :class:`pyopencl.array.Array`
            Source and destination buffers
        src_origin,dest_origin : int
            Offsets for the start of the copy, in bytes
        shape : sequence of int
            Shape of the region to copy (1-3 elements). The first dimension is
            a byte count.
        src_strides,dest_strides : sequence of int
            Strides for the source and destination memory layout, with the same
            length as `shape`. The first element of each must be 1, and each
            element must be a factor of the next element.
        """
        assert src_strides[0] == 1
        assert dest_strides[0] == 1
        pyopencl.enqueue_copy(
            self._pyopencl_command_queue,
            src=src_buffer.data, dest=dest_buffer.data,
            src_origin=(src_origin,), dst_origin=(dest_origin,),
            region=shape,
            src_pitches=src_strides[1:], dst_pitches=dest_strides[1:])

    def enqueue_read_buffer_rect(
            self, buffer, data, buffer_origin, data_origin, shape,
            buffer_strides, data_strides, blocking=True):
        """Copy a region of a buffer to host memory. This is a low-level
        interface that ignores the shape, strides etc of the buffers, and
        treats them as byte arrays. It also only supports 3 or fewer dimensions. Use
        :py:meth:`~katsdpsigproc.accel.DeviceArray.set_region` for a high-level
        interface.

        Parameters
        ----------
        buffer : :class`pyopencl.array.Array`
            Source
        data : array-like
            Target
        buffer_origin, data_origin : int
            Offsets for the start of the copy, in bytes
        shape : sequence of int
            Shape of the region to copy (1-3 elements). The first dimension is
            a byte count.
        buffer_strides,data_strides : sequence of int
            Strides for the destination and source memory layout, with the same
            length as `shape`. The first element of each must be 1, and each
            element must be a factor of the next element.
        blocking : bool, optional
            If true, block until the transfer is complete.
        """
        assert buffer_strides[0] == 1
        assert data_strides[0] == 1
        if hasattr(data, 'enqueue_read_buffer_rect'):
            data.enqueue_read_buffer_rect(self, buffer, buffer_origin, data_origin, shape,
                                          buffer_strides, data_strides, blocking)
        else:
            pyopencl.enqueue_copy(
                self._pyopencl_command_queue,
                src=buffer.data, dest=data,
                host_origin=(data_origin,), buffer_origin=(buffer_origin,),
                region=shape,
                host_pitches=data_strides[1:], buffer_pitches=buffer_strides[1:])

    def enqueue_write_buffer_rect(
            self, buffer, data, buffer_origin, data_origin, shape,
            buffer_strides, data_strides, blocking=True):
        """Copy a region of host memory to a buffer. This is a low-level
        interface that ignores the shape, strides etc of the buffers, and
        treats them as byte arrays. It also only supports 3 or fewer dimensions. Use
        :py:meth:`~katsdpsigproc.accel.DeviceArray.set_region` for a high-level
        interface.

        Parameters
        ----------
        buffer : :class`pyopencl.array.Array`
            Target
        data : array-like
            Source
        buffer_origin, data_origin : int
            Offsets for the start of the copy, in bytes
        shape : sequence of int
            Shape of the region to copy (1-3 elements). The first dimension is
            a byte count.
        buffer_strides,data_strides : sequence of int
            Strides for the destination and source memory layout, with the same
            length as `shape`. The first element of each must be 1, and each
            element must be a factor of the next element.
        blocking : bool, optional
            If true, block until the transfer is complete.
        """
        assert buffer_strides[0] == 1
        assert data_strides[0] == 1
        if hasattr(data, 'enqueue_write_buffer_rect'):
            data.enqueue_write_buffer_rect(self, buffer, buffer_origin, data_origin, shape,
                                           buffer_strides, data_strides, blocking)
        else:
            pyopencl.enqueue_copy(
                self._pyopencl_command_queue,
                src=data, dest=buffer.data,
                host_origin=(data_origin,), buffer_origin=(buffer_origin,),
                region=shape,
                host_pitches=data_strides[1:], buffer_pitches=buffer_strides[1:])

    def enqueue_zero_buffer(self, buffer):
        """Fill a buffer with zero bytes.

        Parameters
        ----------
        buffer : :class:`pyopencl.array.Array`
        """
        pyopencl.enqueue_fill_buffer(self._pyopencl_command_queue, buffer.data,
                                     np.uint8(0), 0, buffer.data.get_info(pyopencl.mem_info.SIZE))

    @classmethod
    def _raw_arg(cls, arg):
        if isinstance(arg, pyopencl.array.Array):
            return arg.data
        else:
            return arg

    def _enqueue_kernel(self, kernel, args, global_size, local_size):
        """Enqueue a kernel to the command queue.

        This version returns the OpenCL event object. Refer to
        :meth:`enqueue_queue` for the public interface.
        """
        # PyOpenCL doesn't allow Array objects to be passed through
        args = [self._raw_arg(x) for x in args]
        # OpenCL allows local_size to be None, but this is not portable to
        # CUDA
        assert local_size is not None
        assert len(local_size) == len(global_size)
        return kernel._pyopencl_kernel(self._pyopencl_command_queue,
                                       global_size, local_size, *args)

    def enqueue_kernel(self, kernel, args, global_size, local_size):
        """Enqueue a kernel to the command queue.

        .. warning:: It is not thread-safe to call this function in two threads
            on the same kernel at the same time.

        Parameters
        ----------
        kernel : :class:`Kernel`
            Kernel to run
        args : sequence
            Arguments to pass to the kernel. Refer to the PyOpenCL
            documentation for details. Additionally, this function allows
            a :class:`pyopencl.array.Array` to be passed.
        global_size : tuple
            Number of work-items in each global dimension
        local_size : tuple
            Number of work-items in each local dimension. Must divide
            exactly into `global_size`.
        """
        self._enqueue_kernel(kernel, args, global_size, local_size)

    def enqueue_marker(self):
        """Create an event at this point in the command queue"""
        return Event(pyopencl.enqueue_marker(self._pyopencl_command_queue))

    def enqueue_wait_for_events(self, events):
        """Enqueue a barrier to wait for all events in `events`."""
        # OpenCL has some odd semantics for an empty wait list, hence the check
        if events:
            pyopencl.enqueue_barrier(self._pyopencl_command_queue,
                                     [x._pyopencl_event for x in events])

    def flush(self):
        """Start enqueued work running, but do not wait for it to complete"""
        self._pyopencl_command_queue.flush()

    def finish(self):
        """Block until all enqueued work has completed"""
        self._pyopencl_command_queue.finish()


class TuningCommandQueue(CommandQueue):
    """Command queue with extra facilities for autotuning. It keeps
    track of kernels that are enqueued since the last call to
    :meth:`start_tuning`, and reports the total time they consume
    when :meth:`stop_tuning` is called.
    """

    def __init__(self, *args, **kwargs):
        kwargs['profile'] = True
        super(TuningCommandQueue, self).__init__(*args, **kwargs)
        self.is_tuning = False
        self.events = []

    def start_tuning(self):
        self.is_tuning = True
        self.events = []

    def enqueue_kernel(self, kernel, args, global_size, local_size):
        if not self.is_tuning:
            super(TuningCommandQueue, self).enqueue_kernel(kernel, args, global_size, local_size)
        else:
            event = self._enqueue_kernel(kernel, args, global_size, local_size)
            self.events.append(event)

    def stop_tuning(self):
        self.finish()
        elapsed = 0.0
        if self.events:
            start = self.events[0].profile.start
            end = self.events[-1].profile.end
            elapsed = (end - start) * 1e-9
        self.is_tuning = False
        self.events = []
        return elapsed



"""Tests for rank.mako"""

import numpy as np
from nose.tools import assert_equal

from .. import accel
from ..accel import DeviceArray, build
from .test_accel import device_test


@device_test
def setup(context, queue):
    global _program, _wgs, _data, _expected, _N, _M
    _wgs = 128
    _M = 1000
    _N = 2000
    _program = build(context, 'test/test_rank.mako', {'size': _wgs})
    rs = np.random.RandomState(seed=1)   # Fixed seed to make test repeatable
    _data = rs.randint(0, _M, size=_N).astype(np.int32)
    _expected = np.empty(_M, dtype=np.int32)
    for i in range(_M):
        _expected[i] = np.sum(np.less(_data, i))


def check_rank(context, queue, kernel_name, wgs):
    data_d = DeviceArray(context, shape=_data.shape, dtype=_data.dtype)
    data_d.set(queue, _data)
    out_d = DeviceArray(context, shape=_expected.shape, dtype=_expected.dtype)
    kernel = _program.get_kernel(kernel_name)
    queue.enqueue_kernel(
        kernel, [
            data_d.buffer,
            out_d.buffer,
            np.int32(_N),
            np.int32(_M)
        ],
        global_size=(wgs,), local_size=(wgs,))
    out = out_d.get(queue)
    np.testing.assert_equal(_expected, out)


@device_test
def test_rank_serial(context, queue):
    check_rank(context, queue, 'test_rank_serial', 1)


@device_test
def test_rank_parallel(context, queue):
    check_rank(context, queue, 'test_rank_parallel', _wgs)


def run_float_func(context, queue, kernel_name, data, output_size):
    """Common code for testing find_min_float, find_max_float, median_non_zero_float"""
    size = 128  # Number of workitems
    kernel = _program.get_kernel(kernel_name)
    data = np.asarray(data, dtype=np.float32)
    N = data.shape[0]
    data_d = accel.DeviceArray(context, shape=data.shape, dtype=np.float32)
    data_d.set(queue, data)
    out_d = accel.DeviceArray(context, shape=(output_size,), dtype=np.float32)
    queue.enqueue_kernel(
        kernel,
        [data_d.buffer, out_d.buffer, np.int32(N)],
        global_size=(size,),
        local_size=(size,))
    out = out_d.empty_like()
    out_d.get(queue, out)
    return out


class TestFindMinMaxFloat(object):
    def check_array(self, context, queue, data):
        data = np.asarray(data, dtype=np.float32)
        expected = [np.nanmin(data), np.nanmax(data)]
        out = run_float_func(context, queue, 'test_find_min_max_float', data, 2)
        assert_equal(expected[0], out[0])
        assert_equal(expected[1], out[1])

    @device_test
    def test_single(self, context, queue):
        self.check_array(context, queue, [5.3])

    @device_test
    def test_ordered(self, context, queue):
        data = np.sort(np.random.uniform(-10.0, 10.0, 1000))
        self.check_array(context, queue, data)

    @device_test
    def test_nan(self, context, queue):
        self.check_array(context, queue, [-10.0, 5.5, np.nan, -20.0, np.nan])

    @device_test
    def test_all_nan(self, context, queue):
        # Can't use check_array, because np.nanmin warns if all are NaN,
        # and assert_equal doesn't handle NaN
        data = np.array([np.nan, np.nan], dtype=np.float32)
        out = run_float_func(context, queue, 'test_find_min_max_float', data, 2)
        assert np.isnan(out[0])
        assert np.isnan(out[1])

    @device_test
    def test_random(self, context, queue):
        data = np.random.uniform(-10.0, 10.0, 1000)
        self.check_array(context, queue, data)


class TestMedianNonZero(object):
    def check_array(self, context, queue, data):
        data = np.asarray(data, dtype=np.float32)
        expected = np.median(data[data > 0.0])
        out = run_float_func(context, queue, 'test_median_non_zero', data, 2)
        assert_equal(expected, out[0])
        assert_equal(expected, out[1])

    @device_test
    def test_single(self, context, queue):
        self.check_array(context, queue, [5.3])

    @device_test
    def test_even(self, context, queue):
        self.check_array(context, queue, [1.2, 2.3, 3.4, 4.5])

    @device_test
    def test_zeros(self, context, queue):
        self.check_array(context, queue, [0.0, 0.0, 0.0, 1.2, 5.3, 2.4])

    @device_test
    def test_big_even(self, context, queue):
        data = np.random.random_sample(10001) + 0.5
        data[123] = 0.0
        self.check_array(context, queue, data)

    @device_test
    def test_big_odd(self, context, queue):
        data = np.random.random_sample(10000) + 0.5
        data[456] = 0.0
        self.check_array(context, queue, data)



from __future__ import division, print_function, absolute_import
import numpy as np
from .test_accel import device_test, force_autotune
from .. import accel
from .. import transpose


class TestTranspose(object):
    def test_transpose(self):
        yield self.check_transpose, 4, 5
        yield self.check_transpose, 53, 7
        yield self.check_transpose, 53, 81
        yield self.check_transpose, 32, 64

    @classmethod
    def pad_dimension(cls, dim, extra):
        """Modifies `dim` to have at least `extra` padding"""
        newdim = accel.Dimension(dim.size, min_padded_size=dim.size + extra)
        newdim.link(dim)

    @device_test
    def check_transpose(self, R, C, context, queue):
        template = transpose.TransposeTemplate(context, np.float32, 'float')
        fn = template.instantiate(queue, (R, C))
        # Force some padded, to check that stride calculation works
        self.pad_dimension(fn.slots['src'].dimensions[0], 1)
        self.pad_dimension(fn.slots['src'].dimensions[1], 4)
        self.pad_dimension(fn.slots['dest'].dimensions[0], 2)
        self.pad_dimension(fn.slots['dest'].dimensions[1], 3)

        ary = np.random.randn(R, C).astype(np.float32)
        src = fn.slots['src'].allocate(fn.allocator)
        dest = fn.slots['dest'].allocate(fn.allocator)
        src.set_async(queue, ary)
        fn()
        out = dest.get(queue)
        np.testing.assert_equal(ary.T, out)

    @device_test
    @force_autotune
    def test_autotune(self, context, queue):
        """Check that the autotuner runs successfully"""
        transpose.TransposeTemplate(context, np.uint8, 'unsigned char')



import numpy as np

from .test_accel import device_test, force_autotune
from .. import accel
from .. import maskedsum


class TestMaskedSum(object):
    def test_maskedsum(self):
        for use_amplitudes in (False, True):
            yield self.check_maskedsum, 4096, 2, use_amplitudes
            yield self.check_maskedsum, 4096, 4029, use_amplitudes
            yield self.check_maskedsum, 4096, 4030, use_amplitudes
            yield self.check_maskedsum, 4096, 4031, use_amplitudes
            yield self.check_maskedsum, 4096, 4032, use_amplitudes

    @classmethod
    def pad_dimension(cls, dim, extra):
        """Modifies `dim` to have at least `extra` padding"""
        newdim = accel.Dimension(dim.size, min_padded_size=dim.size + extra)
        newdim.link(dim)

    @device_test
    def check_maskedsum(self, R, C, use_amplitudes, context, queue):
        template = maskedsum.MaskedSumTemplate(context, use_amplitudes)
        fn = template.instantiate(queue, (R, C))
        # Force some padding, to check that stride calculation works
        self.pad_dimension(fn.slots['src'].dimensions[0], 1)
        self.pad_dimension(fn.slots['src'].dimensions[1], 4)
        ary = np.random.randn(R, C, 2).astype(np.float32).view(dtype=np.complex64)[..., 0]
        msk = np.ones((R,)).astype(np.float32)
        src = fn.slots['src'].allocate(fn.allocator)
        mask = fn.slots['mask'].allocate(fn.allocator)
        dest = fn.slots['dest'].allocate(fn.allocator)
        src.set_async(queue, ary)
        mask.set_async(queue, msk)
        fn()
        out = dest.get(queue).reshape(-1)
        if use_amplitudes:
            use_ary = np.abs(ary)
        else:
            use_ary = ary
        expected = np.sum(use_ary * msk.reshape(ary.shape[0], 1), axis=0)
        np.testing.assert_allclose(expected, out, rtol=1e-6)

    @device_test
    @force_autotune
    def test_autotune(self, context, queue):
        """Check that the autotuner runs successfully"""
        maskedsum.MaskedSumTemplate(context, False)
        maskedsum.MaskedSumTemplate(context, True)



"""Utilities for scheduling device operations with asyncio."""

import warnings

from ..resource import *    # noqa: F401, F403


warnings.warn("katsdpsigproc.asyncio.resource is deprecated. Use katsdpsigproc.resource instead",
              DeprecationWarning)



"""Tests for RFI thresholding algorithms"""

import numpy as np

from .. import host
from ...test.test_accel import device_test, force_autotune
from .. import device


def setup():
    global _deviations, _spikes
    shape = (117, 273)
    # Use a fixed seed to make the test repeatable
    rs = np.random.RandomState(seed=1)
    # Pick 1/4 of samples to be RFI
    _spikes = rs.random_sample(shape) < 0.25
    _deviations = rs.standard_normal(shape).astype(np.float32) * 10.0
    _deviations[_spikes] += 200.0


def test_ThresholdSimpleHost():
    check_host_class(host.ThresholdSimpleHost, 11.0)


def test_ThresholdSumHost():
    check_host_class(host.ThresholdSumHost, 11.0)


def check_host_class(cls, n_sigma):
    threshold = cls(n_sigma)
    noise = np.repeat(10.0, _deviations.shape[1]).astype(np.float32)
    flags = threshold(_deviations, noise)
    np.testing.assert_equal(flags.astype(np.bool_), _spikes)


class BaseTestDeviceClass(object):
    @device_test
    def test_result(self, context, queue):
        n_sigma = 11.0
        template = self.factory(context)
        th_host = template.host_class(n_sigma)
        th_device = device.ThresholdHostFromDevice(template, queue, n_sigma=n_sigma)
        noise = np.linspace(0.0, 50.0, _deviations.shape[1]).astype(np.float32)
        flags_host = th_host(_deviations, noise)
        flags_device = th_device(_deviations, noise)
        np.testing.assert_equal(flags_host, flags_device)

    @device_test
    @force_autotune
    def test_autotune(self, context, queue):
        self.factory(context)


class TestThresholdSimpleDevice(BaseTestDeviceClass):
    def factory(self, context):
        return device.ThresholdSimpleDeviceTemplate(context, False)


class TestThresholdSimpleDeviceTransposed(BaseTestDeviceClass):
    def factory(self, context):
        return device.ThresholdSimpleDeviceTemplate(context, True)


class TestThresholdSumDevice(BaseTestDeviceClass):
    def factory(self, context):
        return device.ThresholdSumDeviceTemplate(context)



#!/usr/bin/env python
# for nosetest: nosetests katsdpsigproc.test.test_maskedsumabs

from __future__ import division, print_function, absolute_import
import time
import numpy as np
from katsdpsigproc import accel
from katsdpsigproc import maskedsum as msum

context = accel.create_some_context(True)
queue = context.create_command_queue(profile=True)

data = np.random.randn(4000, 5000, 2).astype(np.float32).view(dtype=np.complex64)[..., 0]
mask = np.ones((4000,)).astype(np.float32)

template = msum.MaskedSumTemplate(context, use_amplitudes=True)
msum = template.instantiate(queue, data.shape)
msum.ensure_all_bound()
msum.buffer('src').set(queue, data)
msum.buffer('mask').set(queue, mask)
start_event = queue.enqueue_marker()
msum()
end_event = queue.enqueue_marker()
out = msum.buffer('dest').get(queue)

t0 = time.time()
expected = np.sum(np.abs(data) * mask.reshape(data.shape[0], 1), axis=0).astype(np.float32)
t1 = time.time()
print('gpu:', end_event.time_since(start_event), 'cpu:', t1 - t0)
np.testing.assert_allclose(expected, out.reshape(-1), rtol=1e-6)



#!/usr/bin/env python
from setuptools import setup, find_packages


tests_require = ['nose', 'asynctest']

setup(
    name="katsdpsigproc",
    description="Karoo Array Telescope accelerated signal processing tools",
    author="MeerKAT SDP team",
    author_email="sdpdev+katsdpsigproc@ska.ac.za",
    packages=find_packages(),
    package_data={'': ['*.mako']},
    url="https://github.com/ska-sa/katsdpsigproc",
    setup_requires=["katversion"],
    install_requires=[
        "numpy>=1.10",
        "scipy",
        "pandas",
        "numba>=0.36.1",   # Older versions have bugs in median functions
        "decorator",
        "mako",
        "appdirs"
    ],
    extras_require={
        "CUDA": ["pycuda>=2015.1.3"],
        "OpenCL": ["pyopencl>=2017.2.1"],
        "test": tests_require,
        "doc": ["sphinx>=1.3", "sphinxcontrib-asyncio", "sphinx-rtd-theme"]
    },
    python_requires=">=3.5",
    tests_require=tests_require,
    zip_safe=False,
    use_katversion=True
)



import numpy as np

from .test_accel import device_test, force_autotune
from .. import accel
from .. import fill


class TestFill(object):
    @classmethod
    def pad_dimension(cls, dim, extra):
        """Modifies `dim` to have at least `extra` padding"""
        newdim = accel.Dimension(dim.size, min_padded_size=dim.size + extra)
        newdim.link(dim)

    @device_test
    def test_fill(self, context, queue):
        shape = (75, 63)
        template = fill.FillTemplate(context, np.uint32, 'unsigned int')
        fn = template.instantiate(queue, shape)
        self.pad_dimension(fn.slots['data'].dimensions[0], 5)
        self.pad_dimension(fn.slots['data'].dimensions[1], 10)
        fn.ensure_all_bound()
        data = fn.buffer('data')
        # Do the fill
        fn.set_value(0xDEADBEEF)
        fn()
        # Check the result, including padding
        ret = data.get(queue)
        ret = ret.base
        np.testing.assert_equal(ret, 0xDEADBEEF)

    @device_test
    @force_autotune
    def test_autotune(self, context, queue):
        """Test that autotuner runs successfully"""
        fill.FillTemplate(context, np.uint8, 'unsigned char')



"""Tests for :mod:`katsdpsigproc.resource`."""

import time
import asyncio
import queue
import logging

import asynctest

from nose.tools import (assert_equal, assert_true, assert_false,
                        assert_in, assert_not_in, assert_raises, assert_logs)

from .. import resource


class TestWaitUntil(asynctest.TestCase):
    async def test_result(self):
        """wait_until returns before the timeout if a result is set"""
        future = asyncio.Future(loop=self.loop)
        self.loop.call_later(0.1, future.set_result, 42)
        result = await resource.wait_until(future, self.loop.time() + 1000000, loop=self.loop)
        assert_equal(42, result)

    async def test_already_set(self):
        """wait_until returns if a future has a result set before the call"""
        future = asyncio.Future(loop=self.loop)
        future.set_result(42)
        result = await resource.wait_until(future, self.loop.time() + 1000000, loop=self.loop)
        assert_equal(42, result)

    async def test_exception(self):
        """wait_until rethrows an exception set on the future"""
        future = asyncio.Future(loop=self.loop)
        self.loop.call_later(0.1, future.set_exception, ValueError('test'))
        with assert_raises(ValueError):
            await resource.wait_until(future, self.loop.time() + 1000000, loop=self.loop)

    async def test_timeout(self):
        """wait_until throws `asyncio.TimeoutError` if it times out, and cancels the future"""
        future = asyncio.Future(loop=self.loop)
        with assert_raises(asyncio.TimeoutError):
            await resource.wait_until(future, self.loop.time() + 0.01, loop=self.loop)
        assert_true(future.cancelled())

    async def test_shield(self):
        """wait_until does not cancel the future if it is wrapped in shield"""
        future = asyncio.Future(loop=self.loop)
        with assert_raises(asyncio.TimeoutError):
            await resource.wait_until(asyncio.shield(future),
                                      self.loop.time() + 0.01, loop=self.loop)
        assert_false(future.cancelled())


class DummyEvent(object):
    """Dummy version of katsdpsigproc.accel event, whose wait method just
    sleeps for a small time and then appends the event to a queue.
    """
    def __init__(self, completed):
        self.complete = False
        self.completed = completed

    def wait(self):
        if not self.complete:
            time.sleep(0.1)
            self.completed.put(self)
            self.complete = True


class TestResource(asynctest.TestCase):
    def setUp(self):
        self.completed = queue.Queue()

    async def _run_frame(self, acq, event):
        with acq as value:
            assert_equal(42, value)
            await acq.wait_events()
            acq.ready([event])
            self.completed.put(acq)

    async def test_wait_events(self):
        """Test :meth:`resource.ResourceAcquisition.wait_events`"""
        r = resource.Resource(42, loop=self.loop)
        a0 = r.acquire()
        a1 = r.acquire()
        e0 = DummyEvent(self.completed)
        e1 = DummyEvent(self.completed)
        run1 = asyncio.ensure_future(self._run_frame(a1, e1), loop=self.loop)
        run0 = asyncio.ensure_future(self._run_frame(a0, e0), loop=self.loop)
        await run0
        await run1
        order = []
        try:
            while True:
                order.append(self.completed.get_nowait())
        except queue.Empty:
            pass
        assert_equal(order, [a0, e0, a1])

    async def test_context_manager_exception(self):
        """Test using :class:`resource.ResourceAcquisition` as a context
        manager when an error is raised.
        """
        r = resource.Resource(None, loop=self.loop)
        a0 = r.acquire()
        a1 = r.acquire()
        with assert_raises(RuntimeError):
            with a0:
                await a0.wait_events()
                raise RuntimeError('test exception')
        with assert_raises(RuntimeError):
            with a1:
                await a1.wait_events()
                a1.ready()

    async def test_context_manager_no_ready(self):
        """Test using :class:`resource.ResourceAllocation` as a context
        manager when the user does not call
        :meth:`~resource.ResourceAllocation.ready`.
        """
        with assert_logs('katsdpsigproc.resource', logging.WARNING) as cm:
            r = resource.Resource(None, loop=self.loop)
            a0 = r.acquire()
            with a0:
                pass
        assert_equal(
            cm.output,
            ['WARNING:katsdpsigproc.resource:Resource allocation was not explicitly made ready'])


class TestJobQueue(asynctest.TestCase):
    def setUp(self):
        self.jobs = resource.JobQueue()
        self.finished = [asyncio.Future(loop=self.loop) for i in range(5)]
        self.unfinished = [asyncio.Future(loop=self.loop) for i in range(5)]
        for i, future in enumerate(self.finished):
            future.set_result(i)

    def test_clean(self):
        self.jobs.add(self.finished[0])
        self.jobs.add(self.unfinished[0])
        self.jobs.add(self.finished[1])
        self.jobs.clean()
        assert_equal(2, len(self.jobs._jobs))

    async def _finish(self):
        for i, future in enumerate(self.unfinished):
            await asyncio.sleep(0.02, loop=self.loop)
            future.set_result(i)

    async def test_finish(self):
        self.jobs.add(self.finished[0])
        self.jobs.add(self.unfinished[0])
        self.jobs.add(self.unfinished[1])
        self.jobs.add(self.unfinished[2])
        finisher = asyncio.ensure_future(self._finish(), loop=self.loop)
        await self.jobs.finish(max_remaining=1)
        assert_true(self.unfinished[0].done())
        assert_true(self.unfinished[1].done())
        assert_false(self.unfinished[2].done())
        assert_equal(1, len(self.jobs))
        await finisher

    def test_nonzero(self):
        assert_false(self.jobs)
        self.jobs.add(self.finished[0])
        assert_true(self.jobs)

    def test_len(self):
        assert_equal(0, len(self.jobs))
        self.jobs.add(self.finished[0])
        self.jobs.add(self.unfinished[1])
        self.jobs.add(self.finished[1])
        assert_equal(3, len(self.jobs))

    def test_contains(self):
        assert_not_in(self.finished[0], self.jobs)
        self.jobs.add(self.finished[0])
        assert_in(self.finished[0], self.jobs)
        assert_not_in(self.finished[1], self.jobs)



"""Tests for wg_reduce.mako and reduce.py"""

import numpy as np

from .test_accel import device_test, force_autotune
from ..accel import DeviceArray, build
from .. import reduce


class Fixture(object):
    def __init__(self, context, rs, size, allow_shuffle, broadcast):
        rows = 256 // size
        self.data = rs.randint(0, 1000, size=(rows, size)).astype(np.int32)
        self.broadcast = broadcast
        self.program = build(context, 'test/test_reduce.mako', {
            'size': size,
            'allow_shuffle': allow_shuffle,
            'broadcast': broadcast,
            'rows': rows})
        self._description = "Fixture(..., {}, {}, {})".format(size, allow_shuffle, broadcast)

    def __str__(self):
        return self._description


@device_test
def setup(context, queue):
    global _fixtures
    rs = np.random.RandomState(seed=1)  # Fixed seed to make test repeatable
    _fixtures = [Fixture(context, rs, size, allow_shuffle, broadcast)
                 for size in [1, 4, 12, 16, 32, 87, 97, 160, 256]
                 for allow_shuffle in [True, False]
                 for broadcast in [True, False]]


def check_reduce(context, queue, kernel_name, op):
    for fixture in _fixtures:
        data = fixture.data
        kernel = fixture.program.get_kernel(kernel_name)
        data_d = DeviceArray(context=context, shape=data.shape, dtype=data.dtype)
        data_d.set(queue, data)
        out_d = DeviceArray(context=context, shape=data.shape, dtype=data.dtype)
        dims = tuple(reversed(data.shape))
        queue.enqueue_kernel(
            kernel, [data_d.buffer, out_d.buffer], local_size=dims, global_size=dims)
        out = out_d.get(queue)
        expected = op.reduce(data, axis=1, keepdims=True)
        if fixture.broadcast:
            expected = np.repeat(expected, data.shape[1], axis=1)
        else:
            out = out[:, 0:1]
        np.testing.assert_array_equal(expected, out)


@device_test
def test_reduce_add(context, queue):
    check_reduce(context, queue, 'test_reduce_add', np.add)


@device_test
def test_reduce_max(context, queue):
    check_reduce(context, queue, 'test_reduce_max', np.maximum)


class TestHReduce(object):
    """Tests for :class:`katsdpsigproc.reduce.HReduce`"""

    def check(self, context, queue, rows, columns, column_range):
        template = reduce.HReduceTemplate(context, np.uint32, 'unsigned int', 'a + b', 0)
        fn = template.instantiate(queue, (rows, columns), column_range)
        fn.ensure_all_bound()
        device_src = fn.buffer('src')
        device_dest = fn.buffer('dest')
        src = device_src.empty_like()
        rs = np.random.RandomState(seed=1)  # To be reproducible
        src[:] = rs.randint(0, 100000, (rows, columns))
        device_src.set(queue, src)
        fn()
        dest = device_dest.get(queue)
        expected = np.sum(src[:, column_range[0]:column_range[1]], axis=1)
        np.testing.assert_equal(expected, dest)

    @device_test
    def test_normal(self, context, queue):
        """Test the usual case"""
        self.check(context, queue, 129, 173, (67, 128))

    @device_test
    def test_small(self, context, queue):
        """Test that the special case for columns < work group size works"""
        self.check(context, queue, 7, 8, (1, 7))

    @device_test
    @force_autotune
    def test_autotune(self, context, queue):
        """Test that autotuner runs successfully"""
        reduce.HReduceTemplate(context, np.uint32, 'unsigned int', 'a + b', '0')



"""Tests for RFI flagging wrappers. The backgrounding and thresholding have
separate tests - this module just tests that they can be glued together
properly."""

import numpy as np

from .. import host
from ...test.test_accel import device_test
from .. import device


def setup():
    global _vis, _spikes, _input_flags
    shape = (117, 131)
    # Use a fixed seed to make the test repeatable
    rs = np.random.RandomState(seed=1)
    _vis = rs.standard_normal(shape) + 1j * rs.standard_normal(shape)
    # Pick 1/16 of samples to be RFI
    # Give them random amplitude from a range, and random phase
    _spikes = rs.random_sample(shape) < 1.0 / 16.0
    _spikes = _spikes.astype(np.uint8)
    rfi_amp = rs.random_sample(shape) * 20.0 + 50.0
    rfi_phase = rs.random_sample(shape) * (2j * np.pi)
    rfi = rfi_amp * np.exp(rfi_phase)
    _vis += _spikes * rfi
    _vis = _vis.astype(np.complex64)
    _input_flags = (rs.random_sample(shape) < 1.0 / 16.0).astype(np.uint8) * 2


def test_flagger_host():
    background = host.BackgroundMedianFilterHost(13)
    noise_est = host.NoiseEstMADHost()
    threshold = host.ThresholdSimpleHost(11.0)
    flagger = host.FlaggerHost(background, noise_est, threshold)
    flags = flagger(_vis)
    np.testing.assert_equal(_spikes, flags)
    # Test again, this time with channel flags
    flags = flagger(_vis, _input_flags[:, 0])
    input_flags = np.broadcast_to(_input_flags[:, 0:1], _vis.shape)
    expected = np.where(input_flags, 0, _spikes)
    np.testing.assert_equal(expected, flags)
    # And again with full input flags
    flags = flagger(_vis, _input_flags)
    expected = np.where(_input_flags, 0, _spikes)
    np.testing.assert_equal(expected, flags)


def check_flagger_device(use_flags, transpose_noise_est, transpose_threshold, context, queue):
    background = device.BackgroundMedianFilterDeviceTemplate(context, 13, use_flags=use_flags)
    if transpose_noise_est:
        noise_est = device.NoiseEstMADTDeviceTemplate(context, 1024)
    else:
        noise_est = device.NoiseEstMADDeviceTemplate(context, tuning={'wgsx': 8, 'wgsy': 8})
    threshold = device.ThresholdSimpleDeviceTemplate(
        context, transpose_threshold, tuning={'wgsx': 8, 'wgsy': 8})
    flagger_device = device.FlaggerDeviceTemplate(background, noise_est, threshold)
    flagger = device.FlaggerHostFromDevice(flagger_device, queue, threshold_args=dict(n_sigma=11.0))
    if use_flags == device.BackgroundFlags.CHANNEL:
        flags = flagger(_vis, _input_flags[:, 0])
        input_flags = np.broadcast_to(_input_flags[:, 0:1], _vis.shape)
        expected = np.where(input_flags, 0, _spikes)
        np.testing.assert_equal(expected, flags)
    elif use_flags == device.BackgroundFlags.FULL:
        flags = flagger(_vis, _input_flags)
        expected = np.where(_input_flags, 0, _spikes)
        np.testing.assert_equal(expected, flags)
    else:
        flags = flagger(_vis)
        np.testing.assert_equal(_spikes, flags)


@device_test
def test_flagger_device(context, queue):
    check_flagger_device(device.BackgroundFlags.NONE, False, False, context, queue)


@device_test
def test_flagger_device_transpose_noise_est(context, queue):
    """Test device flagger with a transposed noise estimator"""
    check_flagger_device(device.BackgroundFlags.CHANNEL, True, False, context, queue)


@device_test
def test_flagger_device_transpose_threshold(context, queue):
    """Test device flagger with a transposed thresholder"""
    check_flagger_device(device.BackgroundFlags.FULL, False, True, context, queue)


@device_test
def test_flagger_device_transpose_both(context, queue):
    """Test device flagger with a transposed noise estimator and thresholder"""
    check_flagger_device(device.BackgroundFlags.NONE, True, True, context, queue)



"""Fill device array with a constant value"""

import numpy as np

from . import accel
from . import tune


class FillTemplate(object):
    """
    Fills a device array with a constant value. The pad elements are also
    filled with this value.

    .. note::
        To fill with zeros, use :meth:`katsdpsigproc.DeviceArray.zero`

    Parameters
    ----------
    context : |Context|
        Context for which kernels will be compiled
    dtype : numpy dtype
        Type of data elements
    ctype : str
        Type (in C/CUDA, not numpy) of data elements
    tuning : dict, optional
        Kernel tuning parameters; if omitted, will autotune. The possible
        parameters are

        - wgs: number of workitems per workgroup
    """

    def __init__(self, context, dtype, ctype, tuning=None):
        self.context = context
        self.dtype = np.dtype(dtype)
        self.ctype = ctype
        if tuning is None:
            tuning = self.autotune(context, dtype, ctype)
        self.wgs = tuning['wgs']
        self.program = accel.build(context, "fill.mako", {
            'wgs': self.wgs,
            'ctype': ctype
        })

    @classmethod
    @tune.autotuner(test={'wgs': 128})
    def autotune(cls, context, dtype, ctype):
        queue = context.create_tuning_command_queue()
        shape = (1048576,)
        data = accel.DeviceArray(context, shape, dtype=dtype)

        def generate(wgs):
            fn = cls(context, dtype, ctype, {'wgs': wgs}).instantiate(queue, shape)
            fn.bind(data=data)
            return tune.make_measure(queue, fn)

        return tune.autotune(generate, wgs=[64, 128, 256, 512])

    def instantiate(self, *args, **kwargs):
        return Fill(self, *args, **kwargs)


class Fill(accel.Operation):
    """Concrete instance of :class:`FillTemplate`.

    .. rubric:: Slots

    **data**
        Array to be filled (padding will be filled too)

    Parameters
    ----------
    template : :class:`FillTemplate`
        Operation template
    command_queue : |CommandQueue|
        Command queue for the operation
    shape : tuple of int
        Shape for the data slot
    allocator : :class:`DeviceAllocator` or :class:`SVMAllocator`, optional
        Allocator used to allocate unbound slots
    """

    def __init__(self, template, command_queue, shape, allocator=None):
        super(Fill, self).__init__(command_queue, allocator)
        self.template = template
        self.kernel = template.program.get_kernel("fill")
        self.shape = shape
        self.slots['data'] = accel.IOSlot(shape, self.template.dtype)
        self.value = self.template.dtype.type()

    def set_value(self, value):
        self.value = self.template.dtype.type(value)

    def _run(self):
        data = self.buffer('data')

        elements = int(np.product(data.padded_shape))
        global_size = accel.roundup(elements, self.template.wgs)
        self.command_queue.enqueue_kernel(
            self.kernel,
            [data.buffer, np.uint32(elements), self.value],
            global_size=(global_size,),
            local_size=(self.template.wgs,))

    def parameters(self):
        return {
            'dtype': self.template.dtype,
            'ctype': self.template.ctype,
            'shape': self.slots['data'].shape,
            'value': self.value
        }
